package imm.se.vending.model;

import java.util.HashMap;
import java.util.Map;
import java.util.Observable;

/**
 * This class represents the controller of a vending machine. It registers the
 * the input of money and button presses, like "cancel" or selecting a fruit.
 * 
 * @author Hubert
 *
 */
public class VendingMachine extends Observable {

	static final VendingMachineState IDLE_STATE = new IdleState();
	static final VendingMachineState BANANA_SELECTED_STATE = new FruitSelectedState(Fruit.BANANA);
	static final VendingMachineState APPLE_SELECTED_STATE = new FruitSelectedState(Fruit.APPLE);

	private Map<Fruit, Integer> fruits = new HashMap<Fruit, Integer>(); // public attribute read-only for public
	private Map<Fruit, Integer> prices = new HashMap<Fruit, Integer>(); // public attribute read-only for public
	private int currentMoney = 0; // public attribute read-only for public
	private int totalMoney = 0; // public attribute read-only for public
	private int rest = 0; // public attribute read-only for public
	private Fruit dispensedItem = null; // public attribute read-only for public

	private VendingMachineState state = new IdleState(); // package visible attribute

	/**
	 * Constructs a new vending machine controller with a specific amount of bananas
	 * and apples.
	 * 
	 * @param bananas
	 *            the number of bananas
	 * @param apples
	 *            the number of apples
	 */
	public VendingMachine(int bananas, int apples) {
		prices.put(Fruit.APPLE, new Integer(3));
		prices.put(Fruit.BANANA, new Integer(7));
		setFruit(Fruit.BANANA, bananas);
		setFruit(Fruit.APPLE, apples);
	}

	/**
	 * Returns the number of fruits of type fruit left that can be sold.
	 * 
	 * @param fruit
	 * @return the number of fruits of type fruit left that can be sold.
	 */
	public int getNumberOfFruit(Fruit fruit) {
		return fruits.get(fruit);
	}

	/**
	 * Simulates putting coins into the machine. Coins don't have a special
	 * representations in this application. Therefore, denomination of the coins are
	 * used, e.g. 5, 2, and 1.
	 * 
	 * @param amount
	 *            the amount of money entered in the machine.
	 */
	public void input(int amount) {
		clearDispenser();
		state.input(this, amount);
	}

	/**
	 * Selects a fruit that is dispensed by the vending machine given that there is
	 * enough money entered to pay for the fruit and that the fruit is available.
	 * 
	 * @param fruit
	 */
	public void selectFruit(Fruit fruit) {
		clearDispenser();
		state.selectFruit(this, fruit);
	}

	/**
	 * Cancels the buying of a fruit. The money entered into the machine is
	 * returned.
	 */
	public void cancel() {
		clearDispenser();
		state.cancel(this);
	}

	/**
	 * Returns the amount of money that was entered after the last fruit was
	 * dispensed.
	 * 
	 * @return
	 */
	public int getCurrentMoney() {
		return currentMoney;
	}

	/**
	 * Returns the amount of money corresponding to how many fruit was sold. These
	 * are the earnings of the owner of the machine.
	 * 
	 * @return
	 */
	public int getTotalMoney() {
		return totalMoney;
	}

	/**
	 * Returns the value of any over payments from the last time a fruit was
	 * dispensed.
	 * 
	 * @return
	 */
	public int getRest() {
		return rest;
	}

	/**
	 * Returns the selected fruit. Null if no fruit is selected.
	 * 
	 * @return
	 */
	public Fruit selectedFruit() {
		return state.selectedFruit();
	}

	/**
	 * Returns the fruit that was last dispensed. Null if no fruit was dispensed.
	 * 
	 * @return the fruit that was last dispensed. Null if no fruit was dispensed.
	 */
	public Fruit getDispensedItem() {
		return dispensedItem;
	}

	/**
	 * Returns the price of a fruit
	 * 
	 * @param fruit
	 * @return
	 */
	public int getPrice(Fruit fruit) {
		return ((Integer) prices.get(fruit)).intValue();
	}

	void setIdleState() {
		setCurrentState(IDLE_STATE);
	}

	void setCurrentStateForFruit(Fruit fruit) {
		switch (fruit) {
		case APPLE:
			setCurrentState(APPLE_SELECTED_STATE);
			break;
		case BANANA:
			setCurrentState(BANANA_SELECTED_STATE);
			break;
		}
	}

	void dispense(Fruit fruit) {
		if (fruit == null) {
			dispense(null, currentMoney);
			return;
		}
		assert hasFruit(fruit);
		setFruit(fruit, getNumberOfFruit(fruit) - 1);
		int price = getPrice(fruit);
		incrementTotalMoney(price);
		dispense(fruit, currentMoney - price);
	}

	private void dispense(Fruit fruit, int rest) {
		setRest(rest);
		setCurrentMoney(0);
		setDispensedItem(fruit);
	}

	private void clearDispenser() {
		setRest(0);
		setDispensedItem(null);
	}

	private void incrementTotalMoney(int price) {
		totalMoney += price;
		setChanged();
		notifyObservers(NotificationType.TOTAL_MONEY);
	}

	boolean hasEnoughMoneyFor(Fruit fruit) {
		return getCurrentMoney() >= getPrice(fruit);
	}

	public boolean hasFruit(Fruit fruit) {
		return getNumberOfFruit(fruit) > 0;
	}

	private void setRest(int i) {
		rest = i;
		setChanged();
		notifyObservers(NotificationType.REST);
	}

	private void setFruit(Fruit fruit, int value) {
		fruits.put(fruit, value);
		setChanged();
		notifyObservers(NotificationType.FRUIT);
	}

	void setCurrentMoney(int i) {
		currentMoney = i;
		setChanged();
		notifyObservers(NotificationType.CURRENT_MONEY);
	}

	void setCurrentState(VendingMachineState s) {
		state = s;
		setChanged();
		notifyObservers(NotificationType.CURRENT_STATE);
	}

	private void setDispensedItem(Fruit dispensedItem) {
		this.dispensedItem = dispensedItem;
		setChanged();
		notifyObservers(NotificationType.DISPENSED_ITEM);
	}

}
